
"""
sqlalchemy breaks this because they reorganized things. Fail.
see: http://www.mail-archive.com/sqlalchemy@googlegroups.com/msg18392.html
  -AMM 6/15/10
"""

try:
    from sqlalchemy.databases.mysql import MySQLDialect
    from sqlalchemy.databases.oracle import OracleDialect
except:
    from sqlalchemy.dialects.mysql.base import MySQLDialect
    from sqlalchemy.dialects.oracle.base import OracleDialect



"""
File       : MSAuth.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: MSAuth class provides auth/authz functionality for micro-services

The auth/authz schema is defined in JSON data-format as list of records, e.g.
[ record1, record2, ...] where individual record has the following structure
{"role": string, "group": string, "service": string, "action": string, "method": [string, string]}
where
- role defines authorization role, e.g. admin
- group defines specific group, e.g. reqmgr
- service identifies MS service, e.g. ms-pileup
- actions refers to a specific API or action, e.g. create
- method contains list of allowed HTTP methods, e.g. ["POST", "PUT"]

For example, here is a possible list of authz rules for MSPileup service

authz_rules = \
    [{"role": "admin", "group": "reqmgr", "service": "ms-pileup", "action": "create", "method": ["POST"]},
     {"role": "production-operator", "group": "dataops", "service": "ms-pileup", "action": "create", "method": ["POST"]},

     {"role": "admin", "group": "reqmgr", "service": "ms-pileup", "action": "delete", "method": ["DELETE"]},
     {"role": "production-operator", "group": "dataops", "service": "ms-pileup", "action": "delete", "method": ["DELETE"]},

     {"role": "admin", "group": "reqmgr", "service": "ms-pileup", "action": "update", "method": ["PUT"]},
     {"role": "production-operator", "group": "dataops", "service": "ms-pileup", "action": "update", "method": ["PUT"]}]
"""

# system modules
import os
import json
from pathlib import Path

# third party modules
import cherrypy

# WMCore modules
from WMCore.MicroService.Tools.Common import getMSLogger
from WMCore.REST.Auth import user_info_from_headers
from Utils.Utilities import lowerCmsHeaders


def readAuthzRules(entry):
    """
    Return authz rules from given entry
    :param entry: either file name or list of authz rules, see docstring of this module
    :return: list of authz rules
    """
    rules = []
    if entry is None:
        return rules
    # of provided is a list data-type it may contaisn our rules
    if isinstance(entry, list) and len(entry) > 0:
        if isinstance(entry[0], dict):
            return entry
    # check if entry is a file name
    if os.path.exists(entry):
        with open(entry, 'r', encoding="utf-8") as fstream:
            rules = json.load(fstream)
            return rules
    msg = f"Unable to read {entry} as authz rules"
    raise Exception(msg)


class MSAuth():
    """
    This class provides auth/authz functionality for micro-services
    """

    def __init__(self, msConfig, **kwargs):
        """
        Provides a basic setup for all the microservices

        :param msConfig: MS service configuration
        :param kwargs: optional parameters
        """
        self.logger = getMSLogger(getattr(msConfig, 'verbose', False), kwargs.get("logger"))
        self.msConfig = msConfig
        self.logger.info("Configuration including default values:\n%s", self.msConfig)

        # load auth/authz configuration
        self.authzRules = readAuthzRules(msConfig.get('authz_rules', None))
        self.authzKey = msConfig['authz_key']  # hmac key of front-end
        # check if provided authz_key is a file name
        if isinstance(self.authzKey, str):
            authFile = Path(self.authzKey)
            if authFile.is_file():
                # msConfig provided file name with authz_key
                with open(authFile, 'rb') as istream:
                    self.authzKey = istream.read()

    def authorizeApiAccess(self, service, action, method=None):
        """
        Check auth role.
        :return: boolean
        """
        # skip authorization on localhost or if no remote address is provided
        # this is only valid if no cms auth headers is set, i.e. in unit tests
        headers = lowerCmsHeaders(cherrypy.request.headers)
        if 'cms-auth-status' not in headers:
            remoteAddr = headers.get('Remote-Addr', '')
            if remoteAddr == '127.0.0.1' or remoteAddr == '':
                return

        # get user information
        user = user_info_from_headers(key=self.authzKey)

        # CMS AUTH headers
        # cms-auth-status, cms-authn, cms-authz, cms-authn-hmac, cms-authn-name, cms-authn-dn
        for entry in self.authzRules:
            # check that user method is allowed
            if method and method not in entry['method']:
                continue
            userMethod = user.get('method', None)
            if userMethod and userMethod != "X509Cert" and userMethod not in entry['method']:
                continue

            # check if entry service and action matches
            if entry['service'] == service and entry['action'] == action:
                # if we have match for service and action we'll check used roles
                # 'roles': {'fake': {'site': set(['fake']), 'group': set(['fake'])}}
                for role, obj in user['roles'].items():
                    if role != entry['role']:
                        continue
                    for group in obj['group']:
                        if group == entry['group']:
                            return

        # if we reach this place we did not match anything and should through exception
        self.logger.error("ERROR: not authorized access to %s method %s", service, action)
        raise cherrypy.HTTPError(403, "You are not authorized to access this resource.")



#!/usr/bin/python
"""
_WMExceptions_

- List of standard exception ids and their
mappings to a human readable message.
"""

WMEXCEPTION = {'WMCORE-1': 'Not allowed to instantiate ',
               'WMCORE-2': 'Problem creating database table ',
               'WMCORE-3': 'Could not find in library class ',
               'WMCORE-4': 'Problem with loading the class ',
               'WMCORE-5': 'Dialect not specified in configuration ',
               'WMCORE-6': 'Message name is same as diagnostic message name! ',
               'WMCORE-7': 'Component name is reserved word. ',
               'WMCORE-8': 'No config section for this component! ',
               'WMCORE-9': 'Problem inserting a trigger flag. ',
               'WMCORE-10': 'Problem setting trigger action. ',
               'WMCORE-11': 'Security exception. ',
               'WMCORE-12': 'Database connection problem ',
               'WMCORE-13': 'Number of retries exceeded'}

"""
WM_JOB_ERROR_CODES
List of job error codes present in WMCore, some of them coming from CMSSW, and a description of the error it represents.

IMPORTANT:
Any change which is done for WM_JOB_ERROR_CODES has to be done also in the following twiki page:
https://twiki.cern.ch/twiki/bin/view/CMSPublic/JobExitCodes
----------------
Range(1 - 512)
    standard ones in Unix and indicate a CMSSW abort that the cmsRun did not catch as exception
----------------
Range(7000 - 9000)
    cmsRun (CMSSW) exit codes. These codes may depend on specific CMSSW version
    https://github.com/cms-sw/cmssw/blob/CMSSW_5_0_1/FWCore/Utilities/interface/EDMException.h#L26
----------------
Range(10000 - 19999)
    Failures related to the environment setup
----------------
Range(50000 - 59999)
    Failures related executable file
----------------
Range(60000 - 69999)
    Failures related staging-OUT
----------------
Range(70000 - 79999)
    Failures related only for WMAgent. (which does not fit into ranges before)
----------------
Range(80000 - 89999)
    Failures related only for CRAB3. (which does not fit into ranges before)
----------------
Range(90000 - 99999)
    Other problems which does not fit to any range before.
----------------
"""

WM_JOB_ERROR_CODES = {-1: "Error return without specification.",
                      1: "Job failed to bootstrap CMSSW; likely a worker node issue.",
                      2: "Interrupt (ANSI).",
                      4: "Illegal instruction (ANSI).",
                      5: "Trace trap (POSIX).",
                      6: "Abort (ANSI) or IOT trap (4.2BSD).",
                      7: "BUS error (4.2BSD).",
                      8: "Floating point exception (ANSI).",
                      9: "killed, unblockable (POSIX) kill -9.",
                      11: "segmentation violation (ANSI) (most likely user application crashed).",
                      15: "Termination (ANSI).",
                      24: "Soft CPU limit exceeded (4.2 BSD).",
                      25: "File size limit exceeded (4.2 BSD).",
                      30: "Power failure restart (System V.).",
                      50: "Required application version not found at the site.",
                      64: "I/O error: cannot open data file (SEAL).",
                      65: "End of job from user application (CMSSW).",
                      66: "Application exception.",
                      70: "Invalid Arguments: Not all the required arguments for cmsRun is passed",
                      71: "Failed to initiate Scram project",
                      72: "Failed to enter Scram project directory",
                      73: "Failed to get Scram runtime",  # Keeping old error here, as not sure from there previously it was taken: "Failed writing to read-only file system"
                      74: "Unable to untar sandbox",
                      81: "Job did not find functioning CMSSW on worker node.",
                      84: "Job failed to open local and fallback files.",
                      85: "Job failed to open local and fallback files.",
                      90: "Application exception.",
                      92: "Job failed to open local and fallback files.",
                      126: "Permission problem or command is not an executable",
                      127: "Command not found",
                      129: "Hangup (POSIX).",
                      132: "Illegal instruction (ANSI).",
                      133: "Trace trap (POSIX).",
                      134: "Abort (ANSI) or IOT trap (4.2 BSD) (most likely user application crashed).",
                      135: "Bus error (4.2 BSD).",
                      137: "SIGKILL; likely an unrelated batch system kill.",
                      139: "Segmentation violation (ANSI).",
                      143: "Termination (ANSI)(or incorporate in the msg text).",
                      147: "Error during attempted file stageout.",
                      151: "Error during attempted file stageout.",
                      152: "CPU limit exceeded (4.2 BSD).",
                      153: "File size limit exceeded (4.2 BSD).",
                      155: "Profiling alarm clock (4.2 BSD).",
                      195: "Job did not produce a FJR; will retry.",
                      243: "Timeout during attempted file stageout.",
                      7000: "Exception from command line processing.",
                      7001: "Configuration File Not Found.",
                      7002: "Configuration File Read Error.",
                      8001: "Other CMS Exception.",
                      8002: "std::exception (other than bad_alloc).",
                      8003: "Unknown Exception.",
                      8004: "std::bad_alloc (memory exhaustion).",
                      8005: "Bad Exception Type (e.g throwing a string).",
                      8006: "ProductNotFound.",
                      8007: "DictionaryNotFound.",
                      8008: "InsertFailure.",
                      8009: "Configuration.",
                      8010: "LogicError.",
                      8011: "UnimplementedFeature.",
                      8012: "InvalidReference.",
                      8013: "NullPointerError.",
                      8014: "NoProductSpecified.",
                      8015: "EventTimeout.",
                      8016: "EventCorruption.",
                      8017: "ScheduleExecutionFailure.",
                      8018: "EventProcessorFailure.",
                      8019: "FileInPathError.",
                      8020: "FileOpenError (Likely a site error).",
                      8021: "FileReadError (May be a site error).",
                      8022: "FatalRootError.",
                      8023: "MismatchedInputFiles.",
                      8024: "ProductDoesNotSupportViews.",
                      8025: "ProductDoesNotSupportPtr.",
                      8026: "NotFound (something other than a product or dictionary not found).",
                      8027: "FormatIncompatibility.",
                      8028: "FileOpenError with fallback.",
                      8029: "The job configuration has no input files specified for secondary input source.",
                      8030: "Exceeded maximum allowed VSize (ExceededResourceVSize).",
                      8031: "Exceeded maximum allowed RSS (ExceededResourceRSS).",
                      8032: "Exceeded maximum allowed time (ExceededResourceTime).",
                      8033: "Could not write output file (FileWriteError) (usually local disk problem).",
                      8034: "FileNameInconsistentWithGUID",
                      8501: "EventGenerationFailure",
                      9000: "cmsRun caught (SIGINT and SIGUSR2) signal.",
                      10031: "Directory VO_CMS_SW_DIR not found. ",  # (CRAB3)
                      10032: "Failed to source CMS Environment setup script such as cmssset_default.sh, grid system or site equivalent script.",  # (CRAB3)
                      10034: "Required application version is not found at the site.",  # (CRAB3)
                      10040: "failed to generate cmsRun cfg file at runtime.",  # (CRAB3)
                      10042: "Unable to stage-in wrapper tarball.",  # (CRAB3)
                      10043: "Unable to bootstrap WMCore libraries (most likely site python is broken).",  # (CRAB3)
                      11001: "Error during job bootstrap: A sandbox must be specified",  # (WMA)
                      11002: "Error during job bootstrap: A job index must be specified",  # (WMA)
                      11003: "Error during job bootstrap: VO_CMS_SW_DIR, OSG_APP, CVMFS  or /cvmfs were not found.",  # (WMA)
                      11004: "Error during job bootstrap: job environment does not contain the init.sh script.",  # (WMA)
                      11005: "Error during job bootstrap: python isn't available on the worker node.",  # (WMA)
                      50110: "Executable file is not found.",  # (WMA)
                      50111: "Executable file has no exe permissions.",  # (WMA)
                      50113: "Executable did not get enough arguments.",  # (CRAB3)
                      50115: "cmsRun did not produce a valid job report at runtime (often means cmsRun segfaulted).",  # (WMA, CRAB3)
                      50116: "Could not determine exit code of cmsRun executable at runtime.",  # (WMA)
                      50513: "Failure to run SCRAM setup scripts.",  # (WMA, CRAB3)
                      50660: "Application terminated by wrapper because using too much RAM (PSS).",  # (WMA, CRAB3)
                      50662: "Application terminated by wrapper because using too much disk.",  # (CRAB3)
                      50664: "Application terminated by wrapper because using too much Wall Clock time.",  # (WMA, CRAB3)
                      50665: "Application terminated by wrapper because it stay idle too long.",  # (CRAB3)
                      50666: "Job removed by condor for unknown reasons.", # (WMA, CRAB3)
                      50669: "Application terminated by wrapper for not defined reason.",  # (CRAB3)
                      60302: "Output file(s) not found.",  # (CRAB3)
                      60307: "General failure during files stage out.",  # (WMA, CRAB3)
                      60311: "Local Stage Out Failure using site specific plugin.",  # (WMA, CRAB3)
                      60312: "Failure in staging in log files during log collection (WMAgent).",  # (WMA)
                      60313: "Failed to clean up any files that were no longer needed (WMAgent).",  # (WMA)
                      60315: "StageOut initialisation error (Due to TFC, SITECONF etc).",  # (WMA)
                      60317: "Forced timeout for stuck stage out.",  # (To be used in CRAB3/ASO)
                      60318: "Internal error in Crab cmscp.py stageout script.",  # (CRAB3)
                      60319: "Failed to do AlcaHarvest stageout.",  # (WMA)
                      60320: "Failure to communicate with ASO server.",  # (CRAB3)
                      60321: "Site related issue: no space, SE down, refused connection.",  # (To be used in CRAB3/ASO)
                      60322: "User is not authorized to write to destination site.",  # (To be used in CRAB3/ASO)
                      60323: "User quota exceeded.",  # (To be used in CRAB3/ASO)
                      60324: "Other stageout exception.",  # (To be used in CRAB3/ASO)
                      60401: "Failure to assemble LFN in direct-to-merge (WMAgent).",  # (WMA)
                      60403: "Timeout during files stage out.",  # (WMA, CRAB3)
                      60404: "Timeout during staging of log archives- status unknown (WMAgent).",  # (WMA)
                      60405: "General failure to stage out log archives (WMAgent).",  # (WMA)
                      60407: "Timeout in staging in log files during log collection (WMAgent).",  # (WMA)
                      60408: "Failure to stage out of log files during log collection (WMAgent).",  # (WMA)
                      60409: "Timeout in stage out of log files during log collection (WMAgent).",  # (WMA)
                      60450: "No output files present in the report",  # (WMA)
                      60451: "Output file lacked adler32 checksum (WMAgent).",  # (WMA)
                      70318: "Failure in DQM upload.",
                      70452: "No run/lumi information in file (WMAgent).",
                      71101: "No sites are available to submit the job because the location of its input(s) do not pass the site whitelist/blacklist restrictions (WMAgent).",
                      71102: "The job can only run at a site that is currently in Aborted state (WMAgent).",
                      71103: "The job can run only at a site that is currently in Draining state (WMAgent).",
                      71104: "JobSubmitter component could not find a job pickle object.",
                      71105: "JobSubmitter component loaded an empty job pickle object.",
                      71300: "The job was killed by the WMAgent, reason is unknown (WMAgent).",
                      71301: "The job was killed by WMAgent because the site it was supposed to run at was set to Aborted (WMAgent).",
                      71302: "The job was killed by WMAgent because the site it was supposed to run at was set to Draining (WMAgent).",
                      71303: "The job was killed by WMAgent because the site it was supposed to run at was set to Down (WMAgent).",
                      71304: "The job was killed by the WMAgent for using too much wallclock time (WMAgent) Job status was Running.",
                      71305: "The job was killed by the WMAgent for using too much wallclock time (WMAgent) Job status was Pending.",
                      71306: "The job was killed by the WMAgent for using too much wallclock time (WMAgent) Job status was Error.",
                      71307: "The job was killed by the WMAgent for using too much wallclock time (WMAgent) Job status was Unkown.",
                      80000: "Internal error in CRAB job wrapper.",  # CRAB3
                      80001: "No exit code set by job wrapper.",  # CRAB3
                      80453: "Unable to determine pset hash from output file (CRAB3).",
                      90000: "Error in CRAB3 post-processing step (includes basically errors in stage out and file metadata upload).",  # CRAB3 TODO: Need to separate. new stageout errors are 60321-60324 and other needed for file metadata.
                      99108: "Skipping this step due to a failure in a previous one.",  # WMA
                      99109: "Uncaught exception in WMAgent step executor.",  # WMA TODO: Maybe move to 7****?
                      99303: "Job failed and the original job report got lost",
                      99304: "Could not find jobCache directory. Job will be failed",
                      99305: "Found single input file with too many events to be processed in a pilot lifetime",
                      99400: "Job is killed by condor_rm or SYSTEM_PERIODIC_REMOVE",
                      99401: "Job is killed by unknown reason ",
                      99996: "Failed to find a step report in the worker node",
                      99997: "Failed to load job report",
                      99998: "Job report with size 0",
                      99999: "Some generic error with the job report"
                     }

# ======================================================================
# Removed error codes and messages which are not used. (Review date: 2015-12-04)
# ======================================================================
# 10001: "LD_LIBRARY_PATH is not defined.",
# 10002: "Failed to setup LCG_LD_LIBRAR_PATH.",
# 10016: "OSG $WORKING_DIR could not be created.",
# 10017: "OSG $WORKING_DIR could not be deleted.",
# 10018: "OSG $WORKING_DIR could not be deleted.",
# 10020: "Shell script cmsset_default.sh to setup cms environment is not found.",
# 10021: "Failed to scram application project using the afs release area.",
# 10022: "Failed to scram application project using CMS sw disribution on the LCG2.",
# 10030: "Middleware not identified.",
# 10033: "Platform is incompatible with the scram version.",
# 10035: "Scram Project Command Failed.", Not used
# 10036: "Scram Runtime Command Failed.",
# 10037: "Failed to find cms_site_config file in software area.",
# 10038: "Failed to find cms_site_catalogue.sh file in software area.",
# 10039: "cms_site_catalogue.sh failed to provide the catalogue.",
# 10041: "fail to find valid client for output stage out.",
# 50112: "User executable shell file is not found.",
# 50114: "OSG $WORKING_DIR could not be deleted.",
# 50117: "Could not update exit code in job report (a variation of 50115).",
# 50663: "Application terminated by wrapper because using too much CPU time.", Not used
# 50700: "Job Wrapper did not produce any usable output file.", Not used
# 50800: "Application segfaulted (likely user code problem).", Not used
# 50998: "Problem calculating file details (i.e. size, checksum etc).", Not used
# 50999: "OSG $WORKING_DIR could not be deleted.", Not used
# 60300: "Either OutputSE or OutputSE_DIR not defined.",
# 60301: "Neither zip nor tar exists.",
# 60303: "File already exists on the SE.",
# 60304: "Failed to create the summary file (production).",
# 60305: "Failed to create a zipped archive (production).",
# 60306: "Failed to copy and register output file.",
# 60308: "An output file was saved to fall back local SE after failing to copy to remote SE.",
# 60309: "Failed to create an output directory in the catalogue.",
# 60310: "Failed to register an output file in the catalogue.",
# 60313: "Failed to delete the output from the previous run via lcg-del command.", # Mh.. only lcg-del? Not used
# 60314: "Failed to invoke ProdAgent StageOut Script.", Not used
# 60316: "Failed to create a directory on the SE.", Not used
# 60317: "Forced timeout for stuck stage out.",  # (To be used in CRAB3/ASO)
# 60402: "Failure to assemble LFN in direct-to-merge by event (WMAgent).", # (WMA)
# 60406: "Failure in staging in log files during log collection (WMAgent).", Not used
# 60410: "Failure in deleting log files in log collection (WMAgent).", # Not used
# 60411: "Timeout in deleting log files in log collection (WMAgent).", # Not used
# 60514: "PreScriptFailure.", # Not used anymore
# 60515: "PreScriptScramFailure.", # Not used anymore
# 60999: "SG $WORKING_DIR could not be deleted.", # Not used
# 70000: "Output_sandbox too big for WMS: output can not be retrieved.", Not used
# 70500: "Warning: problem with ModifyJobReport.", # More details here: https://twiki.cern.ch/twiki/bin/view/CMSPublic/SWGuideCrabFaq#Exit_code_70500 Not used!
# ======================================================================

"""
STAGEOUT_ERRORS
key - exitCode which is also defined in WM_JOB_ERROR_CODES # TODO: to be used for reporting to dashboard and to end users in crab status command
value - is a list, which has dictionaries with the following content:
    regex - Error message which is exposed by gfal-copy/del.
    error-msg - Error msg which will be shown to users in crab status output
    isPermanent - True or False, which is used in CRAB3/ASO for following reasons:
           a) CRAB3 decides should it submit a task to gridScheduler; If it is not permanent,
              task will be submitted, but also error message will be shown in crab status output
           b) CRAB3 Postjob decides should it retry job or not;
           c) ASO decides should it resubmit transfer to FTS;
"""
STAGEOUT_ERRORS = {60317: [{"regex": ".*Cancelled ASO transfer after timeout.*",
                            "error-msg": "ASO Transfer canceled due to timeout.",
                            "isPermanent": True}
                          ],
                   60321: [{"regex": ".*reports could not open connection to.*",
                            "error-msg": "Storage element is not accessible.",
                            "isPermanent": False},
                           {"regex": ".*451 operation failed\\: all pools are full.*",
                            "error-msg": "Destination site does not have enough space on their storage.",
                            "isPermanent": True},
                           {"regex": ".*system error in connect\\: connection refused.*",
                            "error-msg": "Destination site storage refused connection.",
                            "isPermanent": False}
                          ],
                   60322: [{"regex": ".*permission denied.*",
                            "error-msg": "Permission denied.",
                            "isPermanent": True},
                           {"regex": ".*Permission refused.*",
                            "error-msg": "Permission denied.",
                            "isPermanent": True},
                           {"regex": ".*operation not permitted.*",
                            "error-msg": "Operation not allowed.",
                            "isPermanent": True},
                           {"regex": ".*mkdir\\(\\) fail.*",
                            "error-msg": "Can`t create directory.",
                            "isPermanent": True},
                           {"regex": ".*open/create error.*",
                            "error-msg": "Can`t create directory/file on destination site.",
                            "isPermanent": True},
                           {"regex": ".*mkdir\\: cannot create directory.*",
                            "error-msg": "Can`t create directory.",
                            "isPermanent": True},
                           {"regex": ".*530-login incorrect.*",
                            "error-msg": "Permission denied to write to destination site.",
                            "isPermanent": True},
                          ],
                   60323: [{"regex": ".*does not have enough space.*",
                            "error-msg": "User quota exceeded.",
                            "isPermanent": True},
                           {"regex": ".*disk quota exceeded.*",
                            "error-msg": "Disk quota exceeded.",
                            "isPermanent": True},
                           {"regex": ".*HTTP 507.*",
                            "error-msg": "HTTP 507: Disk quota exceeded or Disk full.",
                            "isPermanent": True},
                          ]}



"""
It defines and provides some data structures required by the
Transferor and Monitor threads.
It isn't yet enforced, if any new keys are needed, it's better
to get them added here as well.
"""
from __future__ import division, print_function

# summary metrics for the MSMonitor thread
# (also available through the `info` REST API)
MONITOR_REPORT = dict(start_time=0,
                      end_time=0,
                      execution_time=0,
                      error="",
                      total_num_campaigns=0,
                      total_num_transfers=0,
                      filtered_transfer_docs=0,
                      success_transfer_doc_update=0,
                      failed_transfer_doc_update=0,
                      request_status_updated=0)

# summary metrics for the MSTransferor thread
# (also available through the `info` REST API)
TRANSFEROR_REPORT = dict(start_time=0,
                         end_time=0,
                         execution_time=0,
                         error="",
                         total_num_requests=0,
                         total_num_campaigns=0,
                         nodes_out_of_space=None,
                         success_request_transition=0,
                         failed_request_transition=0,
                         problematic_requests=0,
                         total_num_active_pileups=0,
                         num_datasets_subscribed=0,
                         num_blocks_subscribed=0)

# structure for a transfer document to be written to central CouchDB
TRANSFER_COUCH_DOC = dict(workflowName="",
                          lastUpdate=0,
                          transfers=None)

# structure of a transfer record, it corresponds to the `transfers` field in
# the TRANSFER_COUCH_DOC structure
TRANSFER_RECORD = dict(dataset="",
                       dataType="",
                       transferIDs=set(),
                       campaignName="",
                       completion=[0.0])

# summary metrics for the MSOutput thread
# (also available through the `info` REST API)
OUTPUT_REPORT = dict(thread_id="",
                     start_time=0,
                     end_time=0,
                     execution_time=0,
                     error="",
                     total_num_requests=0,
                     total_num_campaigns=0,
                     num_datasets_subscribed=0,
                     num_data_requests=0)

# summary metrics for the MSPileup thread
# (also available through the `info` REST API)
PILEUP_REPORT = dict(thread_id="",
                     start_time=0,
                     end_time=0,
                     execution_time=0,
                     error="",
                     total_num_pileups=0,
                     total_num_rules=0)

# summary metrics for the MSRuleCleaner thread
# (also available through the `info` REST API)
RULECLEANER_REPORT = dict(thread_id="",
                          start_time=0,
                          end_time=0,
                          execution_time=0,
                          error="",
                          total_num_requests=0,
                          clean_num_requests=0,
                          normal_archived_num_requests=0,
                          force_archived_num_requests=0,
                          success_block_rules_deleted=0,
                          failed_block_rules_deleted=0,
                          success_container_rules_deleted=0,
                          failed_container_rules_deleted=0,
                          success_request_transition=0,
                          failed_request_transition=0)

# summary metrics for the MSUnmerged thread
# (also available through the `info` REST API)
UNMERGED_REPORT = dict(thread_id="",
                       start_time=0,
                       end_time=0,
                       execution_time=0,
                       error="",
                       total_num_rses=0,
                       total_num_files=0,
                       num_rses_cleaned=0,
                       num_files_deleted=0)



#!/usr/bin/env python3
"""
DBSErrors represents generic class to handle DBS Go server errors
"""


import json

from WMCore.Services.DBS.ProdException import ProdException


def formatEx(excepInst):
    """
    _formatEx_

    given a DbdException instance, generate a simple message from it
    """
    msg = "%s:%s %s" % (excepInst.__class__.__name__,
                        excepInst.getErrorMessage(),
                        excepInst.getErrorCode(),
                        )
    return msg


def formatEx3(excepInst):
    """
    _formatEx_

    given a DbdException instance, generate a simple message from it
    """
    msg = "%s:%s" % (excepInst.__class__.__name__, str(excepInst))
    return msg


class DataMgmtError(ProdException):
    """
    _DataMgmtError_

    General Exception from DataMgmt Interface

    """
    def __init__(self, message, errorNo=1000, **data):
        ProdException.__init__(self, message, errorNo, **data)


class DBSWriterError(DataMgmtError):
    """
    _DBSWriterError_

    Generic Exception for DBS Write Error

    """
    def __init__(self, msg, **data):
        DataMgmtError.__init__(self, msg, 1001, **data)


class DBSReaderError(DataMgmtError):
    """
    _DBSReaderError_

    Generic Exception for DBS Read Error

    """
    def __init__(self, msg, **data):
        DataMgmtError.__init__(self, msg, 1002, **data)


class DBSError():
    """
    DBSError provides generic interface for DBS (Go-based) errors
    """
    def __init__(self, ex):
        try:
            if hasattr(ex, "body"):
                # case of DBSClient HTTPError
                self.data = json.loads(ex.body)[0]
            else:
                self.data = json.loads(ex)[0]
        except Exception as exp:
            self.data = str(ex)

    def getHttpCode(self):
        """
        :return: HTTP error code
        """
        if isinstance(self.data, dict):
            return self.data['http']['code']
        return 0

    def getServerCode(self):
        """
        :return: DBS server error code which is defined here
-        https://github.com/dmwm/dbs2go/blob/master/dbs/errors.go
        """
        if isinstance(self.data, dict):
            return self.data['error']['code']
        return 0

    def getMessage(self):
        """
        :return: DBS server error message (consice output, last error in DBS error chain)
        """
        if isinstance(self.data, dict):
            return self.data['error']['message']
        return ""

    def getReason(self):
        """
        :return: DBS server error reason (expanded message)
        """
        if isinstance(self.data, dict):
            return self.data['error']['reason']
        return ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Workflow object representing how Release Validation workflows
have to be dealt with in the input data placement (MSTransferor)
"""

from WMCore.MicroService.MSTransferor.DataStructs.Workflow import Workflow
from WMCore.Services.Rucio.RucioUtils import GROUPING_ALL


class RelValWorkflow(Workflow):
    """
    Class to represent a RelVal workflow in the context
    of input data placement in MSTransferor
    """

    def getRucioGrouping(self):
        """
        Returns the rucio rule grouping for RelVal workflows,
        which must hold all the input data under the same RSE in
        order to harvest full stats.

        :return: a string with the required DID grouping
        """
        return GROUPING_ALL



"""
Set of common utilities for Unified service.

Author: Valentin Kuznetsov <vkuznet [AT] gmail [DOT] com>
Original code: https://github.com/CMSCompOps/WmAgentScripts/Unified
"""

# futures
from __future__ import division, print_function, absolute_import

from future import standard_library
standard_library.install_aliases()

# system modules
import json
import logging
import math
import re
import time
from urllib.parse import quote, unquote

# WMCore modules
from Utils.IteratorTools import grouper
from Utils.CertTools import ckey, cert
from WMCore.Services.pycurl_manager import RequestHandler
from WMCore.Services.pycurl_manager import getdata as multi_getdata

# DBS agregators
from dbs.apis.dbsClient import aggRuns, aggFileLumis


# static variables
STEP_PAT = re.compile(r'Step[0-9]')
TASK_PAT = re.compile(r'Task[0-9]')


def hasHTTPFailed(row):
    """
    Evaluates whether the HTTP request through PyCurl failed or not.

    :param row: dictionary data returned from pycurl_manager module
    :return: a boolean confirming failure or not
    """
    if 'data' not in row:
        return True
    if int(row.get('code', 200)) == 200:
        return False
    return True


def getMSLogger(verbose, logger=None):
    """
    _getMSLogger_

    Return a logger object using the standard WMCore formatter
    :param verbose: boolean setting debug or not
    :return: a logger object
    """
    if logger:
        return logger

    verbose = logging.DEBUG if verbose else logging.INFO
    logger = logging.getLogger()
    logging.basicConfig(format="%(asctime)s:%(levelname)s:%(module)s: %(message)s",
                        level=verbose)
    return logger


def isRelVal(reqDict):
    """
    Helper function to evaluate whether the workflow is RelVal or not.
    :param reqDict: dictionary with the workflow description
    :return: True if it's a RelVal workflow, otherwise False
    """
    return reqDict.get("SubRequestType", "") in ['RelVal', 'HIRelVal']


def dbsInfo(datasets, dbsUrl):
    "Provides DBS info about dataset blocks"
    datasetBlocks = {}
    datasetSizes = {}
    datasetTransfers = {}
    if not datasets:
        return datasetBlocks, datasetSizes, datasetTransfers

    urls = ['%s/blocks?detail=True&dataset=%s' % (dbsUrl, d) for d in datasets]
    logging.info("Executing %d requests against DBS 'blocks' API, with details", len(urls))
    data = multi_getdata(urls, ckey(), cert())

    for row in data:
        dataset = row['url'].split('=')[-1]
        if hasHTTPFailed(row):
            print("FAILURE: dbsInfo for %s. Error: %s %s" % (dataset, row.get('code'), row.get('error')))
            continue
        rows = json.loads(row['data'])
        blocks = []
        size = 0
        datasetTransfers.setdefault(dataset, {})  # flat dict in the format of blockName: blockSize
        for item in rows:
            blocks.append(item['block_name'])
            size += item['block_size']
            datasetTransfers[dataset].update({item['block_name']: item['block_size']})
        datasetBlocks[dataset] = blocks
        datasetSizes[dataset] = size

    return datasetBlocks, datasetSizes, datasetTransfers


def getPileupDocs(mspileupUrl, queryDict):
    """
    Fetch documents from MSPileup according to the query passed in.
    :param mspileupUrl: string with the MSPileup url
    :param queryDict: dictionary with the MongoDB query to run
    :return: returns a list with all the pileup objects, or raises
        an exception in case of failure
    """
    mgr = RequestHandler()
    headers = {'Content-Type': 'application/json'}
    data = mgr.getdata(mspileupUrl, queryDict, headers, verb='POST',
                       ckey=ckey(), cert=cert(), encode=True, decode=True)
    if data and data.get("result", []):
        if "error" in data["result"][0]:
            msg = f"Failed to retrieve MSPileup documents with query: {queryDict}"
            msg += f" and error message: {data}"
            raise RuntimeError(msg)
    return data["result"]


def getPileupDatasetSizes(datasets, phedexUrl):
    """
    Given a list of datasets, find all their blocks with replicas
    available, i.e., blocks that have valid files to be processed,
    and calculate the total dataset size
    :param datasets: list of dataset names
    :param phedexUrl: a string with the PhEDEx URL
    :return: a dictionary of datasets and their respective sizes
    NOTE: Value `None` is returned in case the data-service failed to serve a given request.
    """
    sizeByDset = {}
    if not datasets:
        return sizeByDset

    urls = ['%s/blockreplicas?dataset=%s' % (phedexUrl, dset) for dset in datasets]
    logging.info("Executing %d requests against PhEDEx 'blockreplicas' API", len(urls))
    data = multi_getdata(urls, ckey(), cert())

    for row in data:
        dataset = row['url'].split('=')[-1]
        if row['data'] is None:
            print("Failure in getPileupDatasetSizes for dataset %s. Error: %s %s" % (dataset,
                                                                                     row.get('code'),
                                                                                     row.get('error')))
            sizeByDset.setdefault(dataset, None)
            continue
        rows = json.loads(row['data'])
        sizeByDset.setdefault(dataset, 0)
        try:
            for item in rows['phedex']['block']:
                sizeByDset[dataset] += item['bytes']
        except Exception as exc:
            print("Failure in getPileupDatasetSizes for dataset %s. Error: %s" % (dataset, str(exc)))
            sizeByDset[dataset] = None
    return sizeByDset


def getBlockReplicasAndSize(datasets, phedexUrl, group=None):
    """
    Given a list of datasets, find all their blocks with replicas
    available (thus blocks with at least 1 valid file), completed
    and subscribed.
    If PhEDEx group is provided, make sure it's subscribed under that
    same group.
    :param datasets: list of dataset names
    :param phedexUrl: a string with the PhEDEx URL
    :param group: optional PhEDEx group name
    :return: a dictionary in the form of:
    {"dataset":
        {"block":
            {"blockSize": 111, "locations": ["x", "y"]}
        }
    }
    NOTE: Value `None` is returned in case the data-service failed to serve a given request.
    """
    dsetBlockSize = {}
    if not datasets:
        return dsetBlockSize

    urls = ['%s/blockreplicas?dataset=%s' % (phedexUrl, dset) for dset in datasets]
    logging.info("Executing %d requests against PhEDEx 'blockreplicas' API", len(urls))
    data = multi_getdata(urls, ckey(), cert())

    for row in data:
        dataset = row['url'].split('=')[-1]
        if row['data'] is None:
            print("Failure in getBlockReplicasAndSize for dataset %s. Error: %s %s" % (dataset,
                                                                                       row.get('code'),
                                                                                       row.get('error')))
            dsetBlockSize.setdefault(dataset, None)
            continue
        rows = json.loads(row['data'])
        dsetBlockSize.setdefault(dataset, {})
        try:
            for item in rows['phedex']['block']:
                block = {item['name']: {'blockSize': item['bytes'], 'locations': []}}
                for repli in item['replica']:
                    if repli['complete'] == 'y' and repli['subscribed'] == 'y':
                        if not group:
                            block[item['name']]['locations'].append(repli['node'])
                        elif repli['group'] == group:
                            block[item['name']]['locations'].append(repli['node'])
                dsetBlockSize[dataset].update(block)
        except Exception as exc:
            print("Failure in getBlockReplicasAndSize for dataset %s. Error: %s" % (dataset, str(exc)))
            dsetBlockSize[dataset] = None
    return dsetBlockSize


def getPileupSubscriptions(datasets, phedexUrl, group=None, percentMin=99):
    """
    Provided a list of datasets, find dataset level subscriptions where it's
    as complete as `percent_min`.
    :param datasets: list of dataset names
    :param phedexUrl: a string with the PhEDEx URL
    :param group: optional string with the PhEDEx group
    :param percent_min: only return subscriptions that are this complete
    :return: a dictionary of datasets and a list of their location.
    NOTE: Value `None` is returned in case the data-service failed to serve a given request.
    """
    locationByDset = {}
    if not datasets:
        return locationByDset

    if group:
        url = "%s/subscriptions?group=%s" % (phedexUrl, group)
        url += "&percent_min=%s&dataset=%s"
    else:
        url = "%s/subscriptions?" % phedexUrl
        url += "percent_min=%s&dataset=%s"
    urls = [url % (percentMin, dset) for dset in datasets]

    logging.info("Executing %d requests against PhEDEx 'subscriptions' API", len(urls))
    data = multi_getdata(urls, ckey(), cert())

    for row in data:
        dataset = row['url'].rsplit('=')[-1]
        if row['data'] is None:
            print("Failure in getPileupSubscriptions for dataset %s. Error: %s %s" % (dataset,
                                                                                      row.get('code'),
                                                                                      row.get('error')))
            locationByDset.setdefault(dataset, None)
            continue
        rows = json.loads(row['data'])
        locationByDset.setdefault(dataset, [])
        try:
            for item in rows['phedex']['dataset']:
                for subs in item['subscription']:
                    locationByDset[dataset].append(subs['node'])
        except Exception as exc:
            print("Failure in getPileupSubscriptions for dataset %s. Error: %s" % (dataset, str(exc)))
            locationByDset[dataset] = None
    return locationByDset


def getBlocksByDsetAndRun(datasetName, runList, dbsUrl):
    """
    Given a dataset name and a list of runs, find all the blocks
    :return: flat list of blocks
    """
    blocks = set()
    if isinstance(runList, set):
        runList = list(runList)

    urls = []
    for runSlice in grouper(runList, 50):
        urls.append('%s/blocks?run_num=%s&dataset=%s' % (dbsUrl, str(runSlice).replace(" ", ""), datasetName))
    logging.info("Executing %d requests against DBS 'blocks' API, with run_num list", len(urls))
    data = multi_getdata(urls, ckey(), cert())

    for row in data:
        dataset = row['url'].rsplit('=')[-1]
        if hasHTTPFailed(row):
            msg = "Failure in getBlocksByDsetAndRun for %s. Error: %s %s" % (dataset,
                                                                             row.get('code'),
                                                                             row.get('error'))
            raise RuntimeError(msg)
        rows = json.loads(row['data'])
        for item in rows:
            blocks.add(item['block_name'])

    return list(blocks)


def getFileLumisInBlock(blocks, dbsUrl, validFileOnly=1):
    """
    Given a list of blocks, find their file run lumi information
    in DBS for up to 10 blocks concurrently
    :param blocks: list of block names
    :param dbsUrl: string with the DBS URL
    :param validFileOnly: integer flag for valid files only or not
    :return: a dict of blocks with list of file/run/lumi info
    """
    runLumisByBlock = {}
    urls = ['%s/filelumis?validFileOnly=%d&block_name=%s' % (dbsUrl, validFileOnly, quote(b)) for b in blocks]
    # limit it to 10 concurrent calls not to overload DBS
    logging.info("Executing %d requests against DBS 'filelumis' API, concurrency limited to 10", len(urls))
    data = multi_getdata(urls, ckey(), cert(), num_conn=10)

    for row in data:
        blockName = unquote(row['url'].rsplit('=')[-1])
        if hasHTTPFailed(row):
            msg = "Failure in getFileLumisInBlock for block %s. Error: %s %s" % (blockName,
                                                                                 row.get('code'),
                                                                                 row.get('error'))
            raise RuntimeError(msg)
        rows = json.loads(row['data'])
        rows = aggFileLumis(rows)  # adjust to DBS Go server output
        runLumisByBlock.setdefault(blockName, [])
        for item in rows:
            runLumisByBlock[blockName].append(item)
    return runLumisByBlock


def findBlockParents(blocks, dbsUrl):
    """
    Helper function to find block parents given a list of block names.
    Return a dictionary in the format of:
    {"child dataset name": {"child block": ["parent blocks"],
                            "child block": ["parent blocks"], ...}}
    NOTE: Value `None` is returned in case the data-service failed to serve a given request.
    """
    parentsByBlock = {}
    urls = ['%s/blockparents?block_name=%s' % (dbsUrl, quote(b)) for b in blocks]
    logging.info("Executing %d requests against DBS 'blockparents' API", len(urls))
    data = multi_getdata(urls, ckey(), cert())
    for row in data:
        blockName = unquote(row['url'].rsplit('=')[-1])
        dataset = blockName.split("#")[0]
        if hasHTTPFailed(row):
            print("Failure in findBlockParents for block %s. Error: %s %s" % (blockName,
                                                                              row.get('code'),
                                                                              row.get('error')))
            parentsByBlock.setdefault(dataset, None)
            continue
        rows = json.loads(row['data'])
        try:
            if dataset in parentsByBlock and parentsByBlock[dataset] is None:
                # then one of the block calls has failed, keep it failed!
                continue
            parentsByBlock.setdefault(dataset, {})
            for item in rows:
                parentsByBlock[dataset].setdefault(item['this_block_name'], set())
                parentsByBlock[dataset][item['this_block_name']].add(item['parent_block_name'])
        except Exception as exc:
            print("Failure in findBlockParents for block %s. Error: %s" % (blockName, str(exc)))
            parentsByBlock[dataset] = None
    return parentsByBlock


def getRunsInBlock(blocks, dbsUrl):
    """
    Provided a list of block names, find their run numbers
    :param blocks: list of block names
    :param dbsUrl: string with the DBS URL
    :return: a dictionary of block names and a list of run numbers
    """
    runsByBlock = {}
    urls = ['%s/runs?block_name=%s' % (dbsUrl, quote(b)) for b in blocks]
    logging.info("Executing %d requests against DBS 'runs' API", len(urls))
    data = multi_getdata(urls, ckey(), cert())
    for row in data:
        blockName = unquote(row['url'].rsplit('=')[-1])
        if hasHTTPFailed(row):
            msg = "Failure in getRunsInBlock for block %s. Error: %s %s" % (blockName,
                                                                            row.get('code'),
                                                                            row.get('error'))
            raise RuntimeError(msg)
        rows = json.loads(row['data'])
        rows = aggRuns(rows) # adjust to DBS Go server output
        runsByBlock[blockName] = rows[0]['run_num']
    return runsByBlock


def getWorkflow(requestName, reqMgrUrl):
    "Get list of workflow info from ReqMgr2 data-service for given request name"
    headers = {'Accept': 'application/json'}
    params = {}
    url = '%s/data/request/%s' % (reqMgrUrl, requestName)
    mgr = RequestHandler()
    res = mgr.getdata(url, params=params, headers=headers, ckey=ckey(), cert=cert())
    data = json.loads(res)
    return data.get('result', [])


def getDetoxQuota(url):
    "Get list of workflow info from ReqMgr2 data-service for given request name"
    headers = {}
    params = {}
    mgr = RequestHandler()
    res = mgr.getdata(url, params=params, headers=headers, ckey=ckey(), cert=cert())
    res = res.split('\n')
    return res


def eventsLumisInfo(inputs, dbsUrl, validFileOnly=0, sumOverLumi=0):
    "Get information about events and lumis for given set of inputs: blocks or datasets"
    what = 'dataset'
    eventsLumis = {}
    if not inputs:
        return eventsLumis
    if '#' in inputs[0]:  # inputs are list of blocks
        what = 'block_name'
    urls = ['%s/filesummaries?validFileOnly=%s&sumOverLumi=%s&%s=%s'
            % (dbsUrl, validFileOnly, sumOverLumi, what, quote(i)) for i in inputs]
    data = multi_getdata(urls, ckey(), cert())
    for row in data:
        data = unquote(row['url'].split('=')[-1])
        if hasHTTPFailed(row):
            print("FAILURE: eventsLumisInfo for %s. Error: %s %s" % (data,
                                                                     row.get('code'),
                                                                     row.get('error')))
            continue
        rows = json.loads(row['data'])
        for item in rows:
            eventsLumis[data] = item
    return eventsLumis


def getEventsLumis(dataset, dbsUrl, blocks=None, eventsLumis=None):
    "Helper function to return number of events/lumis for given dataset or blocks"
    nevts = nlumis = 0
    if blocks:
        missingBlocks = [b for b in blocks if b not in eventsLumis]
        if missingBlocks:
            eLumis = eventsLumisInfo(missingBlocks, dbsUrl)
            eventsLumis.update(eLumis)
        for block in blocks:
            data = eventsLumis[block]
            nevts += data['num_event']
            nlumis += data['num_lumi']
        return nevts, nlumis
    if eventsLumis and dataset in eventsLumis:
        data = eventsLumis[dataset]
        return data['num_event'], data['num_lumi']
    eLumis = eventsLumisInfo([dataset], dbsUrl)
    data = eLumis.get(dataset, {'num_event': 0, 'num_lumi': 0})
    return data['num_event'], data['num_lumi']


def getComputingTime(workflow, eventsLumis=None, unit='h', dbsUrl=None, logger=None):
    "Return computing time per give workflow"
    logger = getMSLogger(verbose=True, logger=logger)
    cput = None

    if 'InputDataset' in workflow:
        dataset = workflow['InputDataset']
        if 'BlockWhitelist' in workflow and workflow['BlockWhitelist']:
            nevts, _ = getEventsLumis(dataset, dbsUrl, workflow['BlockWhitelist'], eventsLumis)
        else:
            nevts, _ = getEventsLumis(dataset, dbsUrl, eventsLumis=eventsLumis)
        tpe = workflow['TimePerEvent']
        cput = nevts * tpe
    elif 'Chain' in workflow['RequestType']:
        base = workflow['RequestType'].replace('Chain', '')
        itask = 1
        cput = 0
        carryOn = {}
        while True:
            t = '%s%d' % (base, itask)
            itask += 1
            if t in workflow:
                task = workflow[t]
                if 'InputDataset' in task:
                    dataset = task['InputDataset']
                    if 'BlockWhitelist' in task and task['BlockWhitelist']:
                        nevts, _ = getEventsLumis(dataset, dbsUrl, task['BlockWhitelist'], eventsLumis)
                    else:
                        nevts, _ = getEventsLumis(dataset, dbsUrl, eventsLumis=eventsLumis)
                elif 'Input%s' % base in task:
                    nevts = carryOn[task['Input%s' % base]]
                elif 'RequestNumEvents' in task:
                    nevts = float(task['RequestNumEvents'])
                else:
                    logger.debug("this is not supported, making it zero cput")
                    nevts = 0
                tpe = task.get('TimePerEvent', 1)
                carryOn[task['%sName' % base]] = nevts
                if 'FilterEfficiency' in task:
                    carryOn[task['%sName' % base]] *= task['FilterEfficiency']
                cput += tpe * nevts
            else:
                break
    else:
        nevts = float(workflow.get('RequestNumEvents', 0))
        feff = float(workflow.get('FilterEfficiency', 1))
        tpe = workflow.get('TimePerEvent', 1)
        cput = nevts / feff * tpe

    if cput is None:
        return 0

    if unit == 'm':
        cput = cput / (60.)
    if unit == 'h':
        cput = cput / (60. * 60.)
    if unit == 'd':
        cput = cput / (60. * 60. * 24.)
    return cput


def sigmoid(x):
    "Sigmoid function"
    return 1. / (1 + math.exp(-x))


def getNCopies(cpuHours, minN=2, maxN=3, weight=50000, constant=100000):
    "Calculate number of copies for given workflow"
    func = sigmoid(-constant / weight)
    fact = (maxN - minN) / (1 - func)
    base = (func * maxN - minN) / (func - 1)
    return int(base + fact * sigmoid((cpuHours - constant) / weight))


def teraBytes(size):
    "Return size in TB (Terabytes)"
    return size / (1000 ** 4)


def gigaBytes(size):
    "Return size in GB (Gigabytes), rounded to 2 digits"
    return round(size / (1000 ** 3), 2)


def elapsedTime(time0, msg='Elapsed time', ndigits=1):
    "Helper function to return elapsed time message"
    msg = "%s: %s sec" % (msg, round(time.time() - time0, ndigits))
    return msg


def getRequest(url, params):
    "Helper function to GET data from given URL"
    mgr = RequestHandler()
    headers = {'Accept': 'application/json'}
    verbose = 0
    if 'verbose' in params:
        verbose = params['verbose']
        del params['verbose']
    data = mgr.getdata(url, params, headers, ckey=ckey(), cert=cert(), verbose=verbose)
    return data


def postRequest(url, params):
    "Helper function to POST request to given URL"
    mgr = RequestHandler()
    headers = {'Accept': 'application/json'}
    verbose = 0
    if 'verbose' in params:
        verbose = params['verbose']
        del params['verbose']
    data = mgr.getdata(url, params, headers, ckey=ckey(), cert=cert(),
                       verb='POST', verbose=verbose)
    return data


def getIO(request, dbsUrl):
    "Get input/output info about given request"
    lhe = False
    primary = set()
    parent = set()
    secondary = set()
    if 'Chain' in request['RequestType']:
        base = request['RequestType'].replace('Chain', '')
        item = 1
        while '%s%d' % (base, item) in request:
            alhe, aprimary, aparent, asecondary = \
                ioForTask(request['%s%d' % (base, item)], dbsUrl)
            if alhe:
                lhe = True
            primary.update(aprimary)
            parent.update(aparent)
            secondary.update(asecondary)
            item += 1
    else:
        lhe, primary, parent, secondary = ioForTask(request, dbsUrl)
    return lhe, primary, parent, secondary


def ioForTask(request, dbsUrl):
    "Return lfn, primary, parent and secondary datasets for given request"
    lhe = False
    primary = set()
    parent = set()
    secondary = set()
    if 'InputDataset' in request:
        datasets = request['InputDataset']
        datasets = datasets if isinstance(datasets, list) else [datasets]
        primary = set([r for r in datasets if r])
    if primary and 'IncludeParent' in request and request['IncludeParent']:
        parent = findParent(primary, dbsUrl)
    if 'MCPileup' in request:
        pileups = request['MCPileup']
        pileups = pileups if isinstance(pileups, list) else [pileups]
        secondary = set([r for r in pileups if r])
    if 'LheInputFiles' in request and request['LheInputFiles'] in ['True', True]:
        lhe = True
    return lhe, primary, parent, secondary


def findParent(datasets, dbsUrl):
    """
    Helper function to find the parent dataset.
    It returns a dictionary key'ed by the child dataset
    NOTE: Value `None` is returned in case the data-service failed to serve a given request.
    """
    parentByDset = {}
    if not datasets:
        return parentByDset

    urls = ['%s/datasetparents?dataset=%s' % (dbsUrl, d) for d in datasets]
    logging.info("Executing %d requests against DBS 'datasetparents' API", len(urls))
    data = multi_getdata(urls, ckey(), cert())

    for row in data:
        dataset = row['url'].split('=')[-1]
        if hasHTTPFailed(row):
            print("Failure in findParent for dataset %s. Error: %s %s" % (dataset,
                                                                          row.get('code'),
                                                                          row.get('error')))
            parentByDset.setdefault(dataset, None)
            continue
        rows = json.loads(row['data'])
        try:
            for item in rows:
                parentByDset[item['this_dataset']] = item['parent_dataset']
        except Exception as exc:
            print("Failure in findParent for dataset %s. Error: %s" % (dataset, str(exc)))
            parentByDset[dataset] = None
    return parentByDset



#!/usr/bin/env python
"""
_RequestInfo_

Class to hold and parse all information related to a given request
"""
# futures
from __future__ import division, print_function
from future.utils import viewitems

# system modules
import datetime
# WMCore modules
from pprint import pformat
from copy import deepcopy
from Utils.IteratorTools import grouper
from Utils.Timers import CodeTimer
from WMCore.DataStructs.LumiList import LumiList
from WMCore.MicroService.MSTransferor.DataStructs.DQMHarvestWorkflow import DQMHarvestWorkflow
from WMCore.MicroService.MSTransferor.DataStructs.GrowingWorkflow import GrowingWorkflow
from WMCore.MicroService.MSTransferor.DataStructs.NanoWorkflow import NanoWorkflow
from WMCore.MicroService.MSTransferor.DataStructs.RelValWorkflow import RelValWorkflow
from WMCore.MicroService.MSTransferor.DataStructs.Workflow import Workflow
from WMCore.MicroService.Tools.PycurlRucio import (getRucioToken, getBlocksAndSizeRucio)
from WMCore.MicroService.Tools.Common import (findBlockParents, getBlocksByDsetAndRun,
                                              getFileLumisInBlock, findParent, getRunsInBlock)
from WMCore.MicroService.MSCore.MSCore import MSCore


def isNanoWorkflow(reqDict):
    """
    Function to parse the request dictionary and decide whether it
    corresponds to a MiniAODSIM to NanoAODSIM workflow.
    :param reqDict: dictionary with the workflow description
    :return: a boolean True if workflow is Nano, False otherwise.
    """
    inputDset = ""
    if reqDict['RequestType'] == "TaskChain":
        inputDset = reqDict["Task1"].get("InputDataset", "")
    elif reqDict['RequestType'] == "StepChain":
        inputDset = reqDict["Step1"].get("InputDataset", "")

    if inputDset.endswith("/MINIAODSIM"):
        return True
    return False


def rsesIntersection(listRSEs):
    """
    Helper function to find the intersection of 2 or more
    sets. E.g., used for workflows with multiple datasets.
    :param listRSEs: a list of lists/sets
    :return: a flat list with unique RSEs
    """
    if not listRSEs:
        return []
    if len(listRSEs) == 1:
        return list(listRSEs[0])
    # otherwise, multiple objects
    rses = set(listRSEs[0])
    for itemList in listRSEs:
        rses = rses.intersection(set(itemList))
    return list(rses)


class RequestInfo(MSCore):
    """
    RequestInfo class provides functionality to access and
    manipulate requests.
    """

    def __init__(self, msConfig, rucioObj, logger):
        """
        Basic setup for this RequestInfo module
        """
        extraArgs = {"skipReqMgr": True, "skipRucio": True}
        super(RequestInfo, self).__init__(msConfig, logger=logger, **extraArgs)

        self.rucio = rucioObj
        self.rucioToken = None
        self.tokenValidity = None
        self.openRunning = self.msConfig["openRunning"]
        self.pileupDocs = []

    def __call__(self, reqRecords, pileupDocs):
        """
        Execute the main logic for input data placement, which depends on:
        1. parsing the workflow and finding the parent dataset name, if any
        2. check the expected locations for pileup dataset(s)
        3. discover all the input data blocks, which are valid
        4. create a map of child to parent blocks
        :param reqRecords: input records
        :param pileupDocs: list with dictionary of pileup objects
        :return: output records. Note that in case of failures, it
            will be a subset of the input records.
        """
        self.pileupDocs = pileupDocs
        # obtain new unified Configuration
        uConfig = self.unifiedConfig()
        if not uConfig:
            self.logger.warning("Failed to fetch the latest unified config. Skipping this cycle")
            return []
        self.logger.info("Going to process %d requests.", len(reqRecords))

        # create a Workflow object representing the request, matching
        # against some specific templates
        workflows = self.classifyWorkflows(reqRecords)

        # check Rucio token validity and renew it if needed
        self.setupRucioToken()
        # Step 1: figure out any possible parent datasets
        self.logger.info("Getting/setting parent datasets for %d requests", len(workflows))
        with CodeTimer("### getParentDatasets", logger=self.logger):
            parentMap = self.getParentDatasets(workflows)
            self.setParentDatasets(workflows, parentMap)

        self.setupRucioToken()
        # Step 2: check if pileup datasets are active and their expected locations
        with CodeTimer("### checkSecondaryData", logger=self.logger):
            workflows = self.setSecondaryData(workflows)

        self.setupRucioToken()
        # Step 3: get final list of valid blocks for both parent and primary data
        # considers run, block and lumi lists
        with CodeTimer("### getInputDataBlocks", logger=self.logger):
            blocksByDset = self.getInputDataBlocks(workflows)
            self.setInputDataBlocks(workflows, blocksByDset)

        # Step 4: compose the final list of parent/child dependency
        with CodeTimer("### getParentChildBlocks", logger=self.logger):
            parentageMap = self.getParentChildBlocks(workflows)
            self.setParentChildBlocks(workflows, parentageMap)

        return workflows

    def classifyWorkflows(self, reqRecords):
        """
        This method classifies the provided workflows into their
        respective MS templates. Making it easier to retrieve
        input data and parameters for input data placement.

        :param reqRecords: list of workflow dictionaries (from ReqMgr2)
        :return: a custom python object for the workflow type
        """
        workflows = []
        for record in reqRecords:
            if record.get("SubRequestType") in ['RelVal', 'HIRelVal']:
                wflow = RelValWorkflow(record['RequestName'], record, logger=self.logger)
            elif record.get("RequestType") == "DQMHarvest":
                wflow = DQMHarvestWorkflow(record['RequestName'], record, logger=self.logger)
            elif record.get("OpenRunningTimeout", 0) > self.openRunning:
                wflow = GrowingWorkflow(record['RequestName'], record, logger=self.logger)
            elif isNanoWorkflow(record):
                wflow = NanoWorkflow(record['RequestName'], record, logger=self.logger)
            else:
                wflow = Workflow(record['RequestName'], record, logger=self.logger)

            workflows.append(wflow)
            msg = f"Processing request: {wflow.getName()}, "
            msg += f"with transferor template: {wflow.__class__.__name__}, "
            msg += f"with campaigns: {wflow.getCampaigns()} and "
            msg += f"input data as:\n{pformat(wflow.getDataCampaignMap())}"
            self.logger.info(msg)
        return workflows

    def setupRucioToken(self):
        """
        Check whether Rucio is enabled and create a new token, or renew it if needed
        :return: a tuple with the token and  its expiration date
        """
        if not self.tokenValidity:
            # a brand new token needs to be created. To be done in the coming lines...
            pass
        elif self.tokenValidity:
            # then check the token lifetime
            dateTimeNow = int(datetime.datetime.utcnow().strftime("%s"))
            timeDiff = self.tokenValidity - dateTimeNow
            if timeDiff > 30 * 60: # 30min
                # current token still valid for a while
                return

        self.rucioToken, self.tokenValidity = getRucioToken(self.msConfig['rucioAuthUrl'],
                                                            self.msConfig['rucioAccount'])

    def _workflowRemoval(self, listOfWorkflows, workflowsToRetry):
        """
        Receives the initial list of workflows and another list of workflows
        that failed somewhere in the MS processing (like fetching information
        from the data-services); and remove those workflows from this cycle of
        the MSTransferor, such that they can be retried in the next cycle.
        :param listOfWorkflows: reference to the list of the initial workflows
        :param workflowsToRetry: list of workflows with missing information
        :return: nothing, the workflow removal is done in place
        """
        for wflow in set(workflowsToRetry):
            self.logger.warning("Removing workflow that failed processing in MSTransferor: %s", wflow.getName())
            listOfWorkflows.remove(wflow)

    def getParentDatasets(self, workflows):
        """
        Given a list of requests, find which requests need to process a parent
        dataset, and discover what the parent dataset name is.
        :return: dictionary with the child and the parent dataset
        """
        retryWorkflows = []
        retryDatasets = []
        datasetByDbs = {}
        parentByDset = {}
        for wflow in workflows:
            if wflow.hasParents():
                datasetByDbs.setdefault(wflow.getDbsUrl(), set())
                datasetByDbs[wflow.getDbsUrl()].add(wflow.getInputDataset())

        for dbsUrl, datasets in viewitems(datasetByDbs):
            self.logger.info("Resolving %d dataset parentage against DBS: %s", len(datasets), dbsUrl)
            # first find out what's the parent dataset name
            parentByDset.update(findParent(datasets, dbsUrl))

        # now check if any of our calls failed; if so, workflow needs to be skipped from this cycle
        # FIXME: isn't there a better way to do this?!?
        for dset, value in viewitems(parentByDset):
            if value is None:
                retryDatasets.append(dset)
        if retryDatasets:
            for wflow in workflows:
                if wflow.hasParents() and wflow.getInputDataset() in retryDatasets:
                    retryWorkflows.append(wflow)
            # remove workflows that failed one or more of the bulk queries to the data-service
            self._workflowRemoval(workflows, retryWorkflows)

        return parentByDset

    def setParentDatasets(self, workflows, parentageMap):
        """
        Set the parent dataset for workflows requiring parents
        """
        for wflow in workflows:
            if wflow.hasParents() and wflow.getInputDataset() in parentageMap:
                wflow.setParentDataset(parentageMap[wflow.getInputDataset()])

    def setSecondaryData(self, workflows):
        """
        Given a list of requests, list all the pileup datasets and check their
        configuration in MSPileup, including their expected locations.
        NOTE: pileup documents contain only active pileup.
        :param workflows: a list of Workflow objects
        :return: list of workflow objects with secondary information set.
        """
        for wflow in workflows:
            self.logger.info("Checking secondary dataset for workflow: %s", wflow.getName())
            puRSEs = []
            for puDset in wflow.getPileupDatasets():
                # if pileup is not found, then it will have an empty location
                wflow.setSecondarySummary(puDset)
                for puDoc in self.pileupDocs:
                    if puDset == puDoc["pileupName"]:
                        wflow.setSecondarySummary(puDset, puDoc["expectedRSEs"])
                        puRSEs.append(puDoc["expectedRSEs"])
                        self.logger.info("  have pileup %s expected at RSEs: %s",
                                         puDset, puDoc['expectedRSEs'])
                        break  # continue with the next pileup dataset
            # set an unique intersection of RSEs for multiple pileups
            wflow.setPURSElist(rsesIntersection(puRSEs))
        return workflows

    def getInputDataBlocks(self, workflows):
        """
        Given a list of requests, list all the primary and parent datasets and, find
        their block sizes and which locations host completed and subscribed blocks
        NOTE it only uses valid blocks (i.e., blocks with at least one replica!)
        :param workflows: a list of Workflow objects
        :return: dictionary with dataset and a few block information
        """
        retryWorkflows = []
        retryDatasets = []
        datasets = set()
        for wflow in workflows:
            for dataIn in wflow.getDataCampaignMap():
                if dataIn['type'] in ["primary", "parent"]:
                    datasets.add(dataIn['name'])

        # fetch all block names and their sizes from Rucio
        self.logger.info("Fetching parent/primary block sizes for %d containers against Rucio: %s",
                         len(datasets), self.msConfig['rucioUrl'])
        blocksByDset = getBlocksAndSizeRucio(datasets, self.msConfig['rucioUrl'], self.rucioToken)

        # now check if any of our calls failed; if so, workflow needs to be skipped from this cycle
        # FIXME: isn't there a better way to do this?!?
        for dsetName in blocksByDset:
            if blocksByDset[dsetName] is None:
                retryDatasets.append(dsetName)
        if retryDatasets:
            for wflow in workflows:
                if wflow.getInputDataset() in retryDatasets or wflow.getParentDataset() in retryDatasets:
                    retryWorkflows.append(wflow)
            # remove workflows that failed one or more of the bulk queries to the data-service
            self._workflowRemoval(workflows, retryWorkflows)
        return blocksByDset

    def setInputDataBlocks(self, workflows, blocksByDset):
        """
        Provided a dictionary structure of dictionary, block name, and a couple of
        block information, set the workflow attributes accordingly.
        """
        retryWorkflows = []
        for wflow in workflows:
            try:
                for dataIn in wflow.getDataCampaignMap():
                    if dataIn['type'] == "primary":
                        newBlockDict = self._handleInputDataInfo(wflow, dataIn['name'],
                                                                 blocksByDset[dataIn['name']])
                        wflow.setPrimaryBlocks(newBlockDict)
                    elif dataIn['type'] == "parent":
                        newBlockDict = self._handleInputDataInfo(wflow, dataIn['name'],
                                                                 blocksByDset[dataIn['name']])
                        wflow.setParentBlocks(newBlockDict)
            except Exception:
                self.logger.error("Workflow: %s will be retried in the next cycle", wflow.getName())
                retryWorkflows.append(wflow)
        # remove workflows that failed one or more of the bulk queries to the data-service
        self._workflowRemoval(workflows, retryWorkflows)

    def _handleInputDataInfo(self, wflow, dset, blocksDict):
        """
        Applies any run/block/lumi list on the primary and parent
        blocks provided.
        It's a convoluted logic, such as:
         1) if there is no run/block/lumi list, just return the initial blocksDict
         2) if it has lumi list, filter runs from it and run block discovery
            given a dataset name and a list of runs
         3) if it has RunWhitelist, run block discovery for a given dataset name
            and a list of runs
         4) if it has only RunBlacklist, discover the run list of all initial blocks
            provided in blocksDict and remove blocks matching only the black list
         5) for the steps above, always check whether the block has replicas
         6) NOW that the block data discovery is completed (considering runs):
           * if LumiList is not enabled, just return the current list of blocks
           * else, fetch file/run/lumi information in bulk of blocks and compare it
           to the LumiList, skipping blocks without a single file that matches it.

        Note that the LumiList check is dealt with in a similar way
        as done in the WorkQueue StartPolicyInterface/getMaskedBlocks:

        :param wflow: the Workflow object
        :param dset: dataset name
        :param blocksDict: dictionary of blocks, their size and location
        :return: dictionary of block names and block size
        """
        finalBlocks = {}
        dbsUrl = wflow.getDbsUrl()
        runAllowedList = wflow.getRunWhitelist()
        runForbiddenList = set(wflow.getRunBlacklist())
        lumiList = wflow.getLumilist()
        # if there is no filter on the input data, simply return it
        if not (lumiList or runAllowedList or runForbiddenList):
            return self._removeZeroSizeBlocks(blocksDict)

        if lumiList:
            # LumiList has precedence over RunWhitelist
            runAllowedList = []
            for run in lumiList.getRuns():
                runAllowedList.append(int(run))
            runAllowedList = list(set(runAllowedList))
        if runAllowedList:
            # Run number 1 is not supported by DBSServer
            if int(runAllowedList[0]) == 1:
                finalBlocks = deepcopy(blocksDict)
            else:
                runAllowedList = list(set(runAllowedList) - runForbiddenList)
                self.logger.info("Fetching blocks matching a list of runs for %s", wflow.getName())
                try:
                    blocks = getBlocksByDsetAndRun(dset, runAllowedList, dbsUrl)
                except Exception as exc:
                    msg = "Failed to retrieve blocks by dataset '%s'and run: %s\n" % (dset, runAllowedList)
                    msg += "Error details: %s" % str(exc)
                    self.logger.error(msg)
                    raise
                for block in blocks:
                    if block in blocksDict:
                        finalBlocks[block] = deepcopy(blocksDict[block])
                    else:
                        self.logger.warning("Dropping block existent in DBS but not in Rucio: %s", block)
        elif runForbiddenList:
            # only run blacklist set
            self.logger.info("Fetching runs in blocks for RunBlacklist for %s", wflow.getName())
            try:
                blockRuns = getRunsInBlock(list(blocksDict), dbsUrl)
            except Exception as exc:
                self.logger.error("Failed to bulk retrieve runs per block. Details: %s", str(exc))
                raise
            for block, runs in viewitems(blockRuns):
                if not set(runs).difference(runForbiddenList):
                    self.logger.info("Dropping block with only blacklisted runs: %s", block)
                elif block in blocksDict:
                    finalBlocks[block] = deepcopy(blocksDict[block])

        if lumiList:
            self.logger.info("Fetching block/lumi information for %d blocks in %s",
                             len(finalBlocks), wflow.getName())
            self.logger.debug("with the following run whitelist: %s", runAllowedList)
            goodBlocks = set()
            # now with a smaller set of blocks in hand, we collect their lumi
            # information and discard any blocks not matching the lumi list
            for blockSlice in grouper(finalBlocks, 10):
                try:
                    blockFileLumis = getFileLumisInBlock(blockSlice, dbsUrl, validFileOnly=1)
                except Exception as exc:
                    self.logger.error("Failed to bulk retrieve run/lumi per block. Details: %s", str(exc))
                    raise
                for block, fileLumis in viewitems(blockFileLumis):
                    for fileLumi in fileLumis:
                        if int(fileLumi['run_num']) not in runAllowedList:
                            continue
                        runNumber = str(fileLumi['run_num'])
                        lumis = fileLumi['lumi_section_num']
                        fileMask = LumiList(runsAndLumis={runNumber: lumis})
                        if lumiList & fileMask:
                            # then it has lumis that we need, keep this block and move on
                            goodBlocks.add(block)
                            break
            # last but not least, drop any blocks that are not in the good list
            for block in list(finalBlocks):
                if block not in goodBlocks:
                    self.logger.info("Dropping block not matching LumiList: %s", block)
                    finalBlocks.pop(block)

        return self._removeZeroSizeBlocks(finalBlocks)

    def _removeZeroSizeBlocks(self, blocksDict):
        """
        Given a dictionary of blocks and their block size and location information,
        return only blocks with >0 bytes of block size (Rucio blocks with no replicas/
        files result in blocks with None size).
        :return: dictionary of block names and block size
        """
        finalBlocks = {}
        for blockName in blocksDict:
            if blocksDict[blockName]['blockSize']:
                finalBlocks[blockName] = blocksDict[blockName]
            else:
                self.logger.info("Dropping block: %s with no files and size: %s",
                                 blockName, blocksDict[blockName]['blockSize'])
        return finalBlocks

    def getParentChildBlocks(self, workflows):
        """
        Given a list of requests, get their children block, discover their parent blocks
        and finally filter out any parent blocks with only invalid files (without any replicas)
        :param workflows: list of workflow objects
        :return: nothing, updates the workflow attributes in place
        """
        retryWorkflows = []
        retryDatasets = []
        blocksByDbs = {}
        parentageMap = {}
        for wflow in workflows:
            blocksByDbs.setdefault(wflow.getDbsUrl(), set())
            if wflow.getParentDataset():
                blocksByDbs[wflow.getDbsUrl()] = blocksByDbs[wflow.getDbsUrl()] | set(wflow.getPrimaryBlocks().keys())

        for dbsUrl, blocks in viewitems(blocksByDbs):
            if not blocks:
                continue
            self.logger.debug("Fetching DBS parent blocks for %d children blocks...", len(blocks))
            # first find out what's the parent dataset name
            parentageMap.update(findBlockParents(blocks, dbsUrl))

        # now check if any of our calls failed; if so, workflow needs to be skipped from this cycle
        # FIXME: isn't there a better way to do this?!?
        for dset, value in viewitems(parentageMap):
            if value is None:
                retryDatasets.append(dset)
        if retryDatasets:
            for wflow in workflows:
                if wflow.getParentDataset() in retryDatasets:
                    retryWorkflows.append(wflow)
            # remove workflows that failed one or more of the bulk queries to the data-service
            self._workflowRemoval(workflows, retryWorkflows)
        return parentageMap

    def setParentChildBlocks(self, workflows, parentageMap):
        """
        Provided a dictionary with the dataset, the child block and a set
        of the parent blocks, set the workflow attribute accordingly
        """
        for wflow in workflows:
            if wflow.getParentDataset() and wflow.getInputDataset() in parentageMap:
                wflow.setChildToParentBlocks(parentageMap[wflow.getInputDataset()])




from Utils.Utilities import encodeUnicodeToBytes
from future.utils import viewitems, viewvalues, listitems

import os, hmac, hashlib, cherrypy
from tempfile import NamedTemporaryFile

from Utils.PythonVersion import PY3

from WMCore.REST.Main import RESTMain
from WMCore.REST.Auth import authz_canonical
from WMCore.Configuration import Configuration

def fake_authz_headers(hmac_key, method = 'HNLogin',
                       login='testuser', name='Test User',
                       dn="/test/dn", roles={}, format="list"):
    """Create fake authentication and authorisation headers compatible
    with the CMSWEB front-ends. Assumes you have the HMAC signing key
    the back-end will use to validate the headers.

    :arg str hmac_key: binary key data for signing headers.
    :arg str method: authentication method, one of X509Cert, X509Proxy,
      HNLogin, HostIP, AUCookie or None.
    :arg str login: account login name.
    :arg str name: account user name.
    :arg str dn: account X509 subject.
    :arg dict roles: role dictionary, each role with 'site' and 'group' lists.
    :returns: list of header name, value tuples to add to a HTTP request."""
    headers = { 'cms-auth-status': 'OK', 'cms-authn-method': method }

    if login:
        headers['cms-authn-login'] = login

    if name:
        headers['cms-authn-name'] = name

    if dn:
        headers['cms-authn-dn'] = dn

    for name, role in viewitems(roles):
        name = 'cms-authz-' + authz_canonical(name)
        headers[name] = []
        for r in 'site', 'group':
            if r in role:
                headers[name].extend(["%s:%s" % (r, authz_canonical(v)) for v in role[r]])
        headers[name] = " ".join(headers[name])

    prefix = suffix = ""
    hkeys = list(headers)
    for hk in sorted(hkeys):
        if hk != 'cms-auth-status':
            prefix += "h%xv%x" % (len(hk), len(headers[hk]))
            suffix += "%s%s" % (hk, headers[hk])

    msg = prefix + "#" + suffix
    if PY3:
        hmac_key = encodeUnicodeToBytes(hmac_key)
        msg = encodeUnicodeToBytes(msg)
    cksum = hmac.new(hmac_key, msg, hashlib.sha1).hexdigest()
    headers['cms-authn-hmac'] = cksum
    if format == "list":
        return listitems(headers)
    else:
        return headers

def fake_authz_key_file(delete=True):
    """Create temporary file for fake authorisation hmac signing key.

    :returns: Instance of :class:`~.NamedTemporaryFile`, whose *data*
      attribute contains the HMAC signing binary key."""
    t = NamedTemporaryFile(delete=delete)
    with open("/dev/urandom", "rb") as fd:
        t.data = fd.read(20)
    t.write(t.data)
    t.seek(0)
    return t

def setup_dummy_server(module_name, class_name, app_name = None, authz_key_file=None, port=8888):
    """Helper function to set up a :class:`~.RESTMain` server from given
    module and class. Creates a fake server configuration and instantiates
    the server application from it.

    :arg str module_name: module from which to import test class.
    :arg str class_type: name of the server test class.
    :arg str app_name: optional test application name, 'test' by default.
    :returns: tuple with the server object and authz hmac signing key."""
    if authz_key_file:
        test_authz_key = authz_key_file
    else:
        test_authz_key = fake_authz_key_file()

    cfg = Configuration()
    main = cfg.section_('main')
    main.application = app_name or 'test'
    main.silent = True
    main.index = 'top'
    main.authz_defaults = { 'role': None, 'group': None, 'site': None }
    main.section_('tools').section_('cms_auth').key_file = test_authz_key.name

    app = cfg.section_(app_name or 'test')
    app.admin = 'dada@example.org'
    app.description = app.title = 'Test'

    views = cfg.section_('views')
    top = views.section_('top')
    top.object = module_name + "." + class_name

    server = RESTMain(cfg, os.getcwd())
    server.validate_config()
    server.setup_server()
    server.install_application()
    cherrypy.config.update({'server.socket_port': port})
    cherrypy.config.update({'server.socket_host': '127.0.0.1'})
    cherrypy.config.update({'request.show_tracebacks': True})
    cherrypy.config.update({'environment': 'test_suite'})
    for app in viewvalues(cherrypy.tree.apps): 
        if '/' in app.config:
            app.config["/"]["request.show_tracebacks"] = True

    return server, test_authz_key



#!/usr/bin/python
"""
_WMException_

General Exception class for WM modules

"""

from builtins import str, bytes
from future.utils import viewitems

import inspect
import logging
import sys
import traceback

from Utils.Utilities import decodeBytesToUnicode
from WMCore.Configuration import PY3

WMEXCEPTION_START_STR = str("<@========== WMException Start ==========@>")
WMEXCEPTION_END_STR = str("<@---------- WMException End ----------@>")


class WMException(Exception):
    """
    _WMException_

    Exception class which works out details of where
    it was raised.

    """

    def __init__(self, message, errorNo=None, **data):
        self.name = str(self.__class__.__name__)

        # Fix for the unicode encoding issue, see #8056 and #8403
        # interprets this string using utf-8 codec and ignoring any errors.
        # using unicode sandwich pattern
        message = decodeBytesToUnicode(message, "ignore")

        Exception.__init__(self, self.name, message)

        #  //
        # // Init data dictionary with defaults
        # //
        self.data = {}
        self.data.setdefault("ClassName", None)
        self.data.setdefault("ModuleName", None)
        self.data.setdefault("MethodName", None)
        self.data.setdefault("ClassInstance", None)
        self.data.setdefault("FileName", None)
        self.data.setdefault("LineNumber", None)
        if errorNo is None:
            self.data.setdefault("ErrorNr", 0)
        else:
            self.data.setdefault("ErrorNr", errorNo)

        self._message = message
        self.addInfo(**data)

        #  //
        # // Automatically determine the module name
        # //  if not set
        if self.data['ModuleName'] is None:
            try:
                frame = inspect.currentframe()
                lastframe = inspect.getouterframes(frame)[1][0]
                excepModule = inspect.getmodule(lastframe)
                if excepModule is not None:
                    modName = excepModule.__name__
                    self.data['ModuleName'] = modName
            finally:
                frame = None

        # //
        # // Find out where the exception came from
        # //
        try:
            stack = inspect.stack(1)[1]
            self.data['FileName'] = stack[1]
            self.data['LineNumber'] = stack[2]
            self.data['MethodName'] = stack[3]
        finally:
            stack = None

        # //
        # // ClassName if ClassInstance is passed
        # //
        try:
            if self.data['ClassInstance'] is not None:
                self.data['ClassName'] = self.data['ClassInstance'].__class__.__name__
        except Exception:
            pass

        # Determine the traceback at time of __init__
        try:
            self.traceback = "\n".join(traceback.format_tb(sys.exc_info()[2]))
        except Exception:
            self.traceback = "WMException error: Couldn't get traceback\n"
        # using unicode sandwich pattern
        self.traceback = decodeBytesToUnicode(self.traceback, "ignore")

    def __getitem__(self, key):
        """
        make exception look like a dictionary
        """
        return self.data[key]

    def __setitem__(self, key, value):
        """
        make exception look like a dictionary
        """
        # using unicode sandwich pattern
        key = decodeBytesToUnicode(key, "ignore")
        value = decodeBytesToUnicode(value, "ignore")
        self.data[key] = value

    def addInfo(self, **data):
        """
        _addInfo_

        Add key=value information pairs to an
        exception instance
        """
        for key, value in viewitems(data):
            # assumption: value is not iterable (list, dict, tuple, ...)
            # using unicode sandwich pattern
            key = decodeBytesToUnicode(key, "ignore")
            value = decodeBytesToUnicode(value, "ignore")
            self.data[key] = value
        return

    def xml(self):
        """create a xml string rep of this exception"""
        strg = "<Exception>\n"
        strg += "<Object>\n"
        strg += "%s\n" % self.name
        strg += "</Object>\n"
        strg += "<Message>\n"
        strg += self._message
        strg += "</Message>\n"
        strg += "<DataItems>\n"
        for key, value in viewitems(self.data):
            strg += "<DataItem>\n"
            strg += "<Key>\n"
            strg += str(key)
            strg += "</Key>\n"
            strg += "<Value>\n"
            strg += str(value)
            strg += "</Value>\n"
            strg += "</DataItem>\n"
        strg += "</DataItems>\n"
        strg += "</Exception>\n"
        logging.error(strg)
        return strg

    def __as_unicode(self):
        """create a string rep of this exception"""
        # WARNING: Do not change this string - it is used to extract error from log
        strg = WMEXCEPTION_START_STR
        strg += "\nException Class: %s\n" % self.name
        strg += "Message: %s\n" % self._message
        for key, value in viewitems(self.data):
            strg += "\t%s : %s\n" % (key, value,)
        strg += "\nTraceback: \n"
        strg += self.traceback
        strg += '\n'
        strg += WMEXCEPTION_END_STR
        return strg

    def __as_bytes(self):
        return self.__as_unicode().encode("utf-8")
    
    if PY3:
        __str__ = __as_unicode
    else:
        __str__ = __as_bytes
    __unicode__ = __as_unicode
    __repr__ = __str__


    def message(self):
        return self._message



#! /usr/bin/env python
"""
Unit testing base class with our extensions
"""

from __future__ import (division, print_function)
from future.utils import viewitems

import copy
import unittest


class ExtendedUnitTestCase(unittest.TestCase):
    """
    Class that can be imported to switch to 'mock'ed versions of
    services.
    """

    def assertContentsEqual(self, expected_obj, actual_obj, msg=None):
        """
        A nested object comparison without regard for the ordering of contents. It asserts that
        expected_obj and actual_obj contain the same elements and that their sub-elements are the same.
        However, all sequences are allowed to contain the same elements, but in different orders.
        """

        def traverse_dict(dictionary):
            for key, value in viewitems(dictionary):
                if isinstance(value, dict):
                    traverse_dict(value)
                elif isinstance(value, list):
                    traverse_list(value)
            return

        def get_dict_sortkey(x):
            if isinstance(x, dict):
                return list(x.keys())
            else:
                return x

        def traverse_list(theList):
            for value in theList:
                if isinstance(value, dict):
                    traverse_dict(value)
                elif isinstance(value, list):
                    traverse_list(value)
            theList.sort(key=get_dict_sortkey)
            return

        if not isinstance(expected_obj, type(actual_obj)):
            self.fail(msg="The two objects are different type and cannot be compared: %s and %s" % (
            type(expected_obj), type(actual_obj)))

        expected = copy.deepcopy(expected_obj)
        actual = copy.deepcopy(actual_obj)

        if isinstance(expected, dict):
            traverse_dict(expected)
            traverse_dict(actual)
        elif isinstance(expected, list):
            traverse_list(expected)
            traverse_list(actual)
        else:
            self.fail(msg="The two objects are different type (%s) and cannot be compared." % type(expected_obj))

        return self.assertEqual(expected, actual)
